/*
** section.c for libelfsh
** 
** Started on  Mon Feb 26 04:12:42 2001 mayhem
** Last update Fri Mar  1 00:02:46 2002 mayhem
*/
#include "libelfsh.h"



/* Return the section file offset */
int	elfsh_get_section_foffset(Elf32_Shdr *s)
{
  return (s->sh_offset);
}

/* Return the section virtual address */
int	elfsh_get_section_vaddr(Elf32_Shdr *s)
{
  return (s->sh_addr);
}

/* Return the alignement for the section */
int	elfsh_get_section_align(Elf32_Shdr *s)
{
  return (s->sh_addralign);
}

/* Return the section size giving the related Elf32_Shdr pointer */
int	elfsh_get_section_size(Elf32_Shdr *s)
{
  return (s->sh_size);
}


/* Lookup the section type giving its sh_type member */
u_int	elfsh_get_section_type(Elf32_Shdr *s)
{
  return (s->sh_type);
}


/* Lookup the name from its index in the string table ;> */
char	*elfsh_get_section_name(Elfsh32_Obj *file, Elf32_Shdr *s)
{
  char	*str;

  if (!s || !file || (!file->sht && !elfsh_get_sht(file, NULL)))
    return (NULL);
  str = file->sh_strtab->data;
  return (str + s->sh_name);
}





/* Return a pointer on a section giving its name */
Elfsh32_Sect	*elfsh_get_section_by_name(Elfsh32_Obj	*file,
					   char		*name,
					   int		*idx,
					   int		*strindex,
					   int		*num)
{
  char		*sname;
  Elfsh32_Sect	*section;
  int		index;

  if (!file->sht && !elfsh_get_sht(file, NULL))
    return (NULL);
  
  for (index = 0, section = file->sections; section; section = section->next, index++)
    {
      sname = elfsh_get_section_name(file, section->shdr);
      if (!strcmp(name, sname))
	{
	  if (idx)
	    *idx = index;
	  if (num)
	    *num = section->shdr->sh_size;
	  if (strindex)
	    *strindex = section->shdr->sh_link;
	  return (section);
	}
    }

  ELFSH_SETERROR("libelfsh: Section not found .", NULL);
}






/* Return the section header for the 'range' occurence of a 'type typed section */
/* Return NULL if failed */
Elfsh32_Sect		*elfsh_get_section_by_type(Elfsh32_Obj	*file,
						   u_int	type,
						   int		*index,
						   int		*strindex,
						   int		*num,
						   int		range)
{
  int			number;
  int			local_index;
  int			nbr;
  Elfsh32_Sect		*section;

  if (!elfsh_get_sht(file, &nbr))
    return (NULL);

  section = file->sections;
  for (local_index = number = 0; local_index < nbr; local_index++)
    {
      if (local_index == file->hdr->e_shnum)
	ELFSH_SETERROR("libelfsh: Section not found .", NULL);
      if (section->shdr->sh_type == type && ++number > range)
	{
	  
	  if (strindex)
	    *strindex = section->shdr->sh_link;
	  
	  if (num)
	    *num = section->shdr->sh_size;  
	  
	  if (index)
	    *index = local_index;

	  return (section);
	  
	}
      section = section->next;
    }
  
  return (NULL);

}






/* Return a Elf32_Shdr pointer giving the section index in the sht */
Elfsh32_Sect		*elfsh_get_section_by_index(Elfsh32_Obj	*file,
						    int		index,
						    int		*strindex,
						    int		*num)
{
  Elfsh32_Sect		*section;
  int			local;

  if (!file->sht && !elfsh_get_sht(file, num))
    return (NULL);
  
  if (strindex)
    *strindex = file->sht[index].sh_link;
  
  if (num)
    *num = file->sht[index].sh_size;
  
  for (section = file->sections, local = 0; local < index; local++)
    section = section->next;

  return (section);
}







/*
** Load the code of a section giving its header
** Should be called after a elfsh_get_section_by_{name,type} to load the code
*/
void	*elfsh_load_section(Elfsh32_Obj	*file, Elf32_Shdr *shdr)
{
  void	*sct;
  
  XSEEK(file->fd, shdr->sh_offset, SEEK_SET, NULL);
  XALLOC(sct, shdr->sh_size, NULL);
  XREAD(file->fd, sct, shdr->sh_size, NULL);
  return (sct);
}



/* Fill an anonymous (unknown content) section */
void	*elfsh_get_anonymous_section(Elfsh32_Obj *file, Elfsh32_Sect *sect)
{
  if (!file || !sect)
    return (NULL);

  if ((sect->data = elfsh_load_section(file, sect->shdr)) == NULL)
    return (NULL);

  sect->nbr = sect->shdr->sh_size;
  return (sect->data);
}





/* Add a given section to the file's section list */
void		elfsh_add_section(Elfsh32_Obj *file, Elfsh32_Sect *sct, int range)
{
  Elfsh32_Sect	*tmp;
  int		index;

  /* in first */
  if (range == ELFSH_SECTION_FIRST)
    {
      if (file->sections)
	{
	  file->sections->prev = sct;
	  sct->next = file->sections;
	}
      file->sections = sct;
    }

  /* in last */
  else if (range == ELFSH_SECTION_LAST)
    {
      if (!file->sections)
	file->sections = sct;
      else
	{
	  tmp = file->sections;
	  while (tmp->next)
	    tmp = tmp->next;
	  tmp->next = sct;
	  sct->prev = tmp;
	}
    }

  /* Rangeth */
  else
    {
      for (tmp = file->sections, index = 0; index < range; tmp = tmp->next)
	index++;
      sct->prev = tmp->prev;
      sct->next = tmp;
      tmp->prev->next = sct;
      tmp->prev = sct;
    }


}




/* Return the parent tht entry for this symbol or NULL if not found */
Elfsh32_Sect	*elfsh_get_parent_section(Elfsh32_Obj *file, u_int value, int *offset)
{
  Elfsh32_Sect	*s;

  if (!file || !value || (!file->sht && !elfsh_get_sht(file, NULL)))
    return (NULL);
  
  for (s = file->sections; s; s = s->next)
    if (INTERVAL(s->shdr->sh_addr, value, (s->shdr->sh_addr + s->shdr->sh_size)))
      {
	if (offset)
	  *offset = value - s->shdr->sh_addr;
	return (s);
      }
  return (NULL);
}




/* Fingerprint the section and fill its type field , NOT IMPLEMENTED YET */
int	elfsh_fingerprint_section(Elfsh32_Sect *sect)
{
  return (ELFSH_SECTION_UNKNOWN);
}





/* Retreive the section header giving the section symbol from .symtab */
Elf32_Shdr		*elfsh_get_section_header_from_sym(Elfsh32_Obj *file, Elf32_Sym *sym)
{
  Elf32_Shdr		*current;
  int			index;
  int			nbr;

  if (!file->sht && !elfsh_get_sht(file, &nbr))
    return (NULL);

  for (index = 0, current = file->sht; index < nbr; index++)
    if (current[index].sh_addr == sym->st_value)
      return (current + index);

  return (NULL);
}









/* Three tiny functions to retreive the rights for a section */
int	elfsh_section_is_allocatable(Elf32_Shdr *s)
{
  return ((s->sh_flags & SHF_ALLOC));
}

int	elfsh_section_is_writable(Elf32_Shdr *s)
{
  return ((s->sh_flags & SHF_WRITE));
}

int	elfsh_section_is_executable(Elf32_Shdr *s)
{
  return ((s->sh_flags & SHF_EXECINSTR));
}








